import numpy as np 

def primes(int nb_primes):
    cdef int n, i, len_p
    cdef int[1000] p
    if nb_primes > 1000:
        nb_primes = 1000
    len_p = 0  # The current number of elements in p.
    n = 2
    while len_p < nb_primes:
        # Is n prime?
        for i in p[:len_p]:
            if n % i == 0:
                break
        # If no break occurred in the loop, we have a prime.
        else:
            p[len_p] = n
            len_p += 1
        n += 1
    # Let's copy the result into a Python list:
    result_as_list = [prime for prime in p[:len_p]]
    return result_as_list


import random
from .engine import Value
from .nn import Module


class RNNNeuron(Module):
 
    def __init__(self, nin, nonlin=True):
        self.w_xh = [Value(random.uniform(-1,1)) for _ in range(nin)]
        self.w_hh = Value(random.uniform(-1,1)) 
        self.b = Value(random.uniform(-1,1))
        self.h = Value(0)
        self.nonlin = nonlin
 
    def __call__(self, x):
        act = sum((wi*xi for wi,xi in zip(self.w_xh, x)), self.b) + (self.w_hh*self.h)
        self.h = act.relu() if self.nonlin else act
        return act.relu() if self.nonlin else act
 
    def parameters(self):
        return self.w_xh + [self.w_hh] + [self.b]
 
    def __repr__(self):
       return f"{'ReLU' if self.nonlin else 'Linear'} RNNNeuron({len(self.w_xh)})"


class RNNLayer(Module):

    def __init__(self, nin, nout, **kwargs):
        self.neurons = [RNNNeuron(nin, **kwargs) for _ in range(nout)]

    def __call__(self, x):
        out = [n(x) for n in self.neurons]
        return out[0] if len(out) == 1 else out

    def parameters(self):
        return [p for n in self.neurons for p in n.parameters()]

    def __repr__(self):
        return f"RNN Layer of [{', '.join(str(n) for n in self.neurons)}]"


class RNN(Module):

    def __init__(self, nin, nouts):
        sz = [nin] + nouts
        self.layers = [RNNLayer(sz[i], sz[i+1], nonlin=i!=len(nouts)-1) for i in range(len(nouts))]

    def __call__(self, x):
        for layer in self.layers:
            x = layer(x)
        return x

    def parameters(self):
        return [p for layer in self.layers for p in layer.parameters()]

    def __repr__(self):
        return f"RNN of [{', '.join(str(layer) for layer in self.layers)}]"


if __name__ == "__main__":

    # NEURON 
    x = Value(1.0)
    print("Neuron forward")
    n = RNNNeuron(1)
    print(n)
    out = n([x])
    print(out)
    print(f"ht = {n.h}")
    print(n.parameters())

    # LAYER
    print("\nLayer forward")
    l = RNNLayer(1, 3)
    print(l)
    out = l([x])
    print(out)
    print(l.parameters())

    # RNN
    print("\nRNN")
    model = RNN(1, [3,1])
    out = model([x])
    print(out)



from .nn import (Neuron, Layer, MLP)
from .rnn import (RNNNeuron, RNNLayer, RNN)
from .graph import (draw_dot)
from .engine import Value
import pyximport
pyximport.install(language_level=3)
from .primes import primes

__version__ = "0.0.2"


import random
from .engine import Value
import pickle


class Module:

    def zero_grad(self):
        for p in self.parameters():
            p.grad = 0

    def parameters(self):
        return []


class Neuron(Module):

    def __init__(self, nin, bias: bool = True, nonlin: bool =True):
        self.w = [Value(random.uniform(-1,1)) for _ in range(nin)]
        if bias: 
            self.b = Value(random.uniform(-1,1))
        else: 
            self.b = Value(0.)
        self.nonlin = nonlin

    def __call__(self, x):
        act = sum((wi*xi for wi,xi in zip(self.w, x)), self.b)
        return act.relu() if self.nonlin else act

    def parameters(self):
        return self.w + [self.b]

    def __repr__(self):
        return f"{'ReLU' if self.nonlin else 'Linear'}Neuron({len(self.w)})"


class Layer(Module):

    def __init__(self, nin, nout, bias: bool = True, **kwargs):
        self.neurons = [Neuron(nin, bias=bias, **kwargs) for _ in range(nout)]

    def __call__(self, x):
        out = [n(x) for n in self.neurons]
        return out[0] if len(out) == 1 else out

    def parameters(self):
        return [p for n in self.neurons for p in n.parameters()]

    def __repr__(self):
        return f"Layer of [{', '.join(str(n) for n in self.neurons)}]"


class MLP(Module):

    def __init__(self, nin, nouts, bias: bool = True):
        sz = [nin] + nouts
        self.layers = [Layer(sz[i], sz[i+1], nonlin=i!=len(nouts)-1, bias = bias) for i in range(len(nouts))]

    def __call__(self, x):
        for layer in self.layers:
            x = layer(x)
        return x

    def parameters(self):
        return [p for layer in self.layers for p in layer.parameters()]
    
    def __repr__(self):
        return f"MLP of [{', '.join(str(layer) for layer in self.layers)}]"
    
    def load(self, save_name: str = "weights"):
        file_name = f'{save_name}.pkl'
        with open(file_name, 'rb') as f:
            params = pickle.load(f)
        load_weights = []
        for param in params:
            if isinstance(param, list):
                load_param = [Value(i) for i in param]
                load_weights.append(load_param)
            else:
                load_param = Value(param)
                load_weights.append(load_param)
        i = 0
        for layer in self.layers:
            for neuron in layer.neurons:
                neuron.w = load_weights[i]
                neuron.b = load_weights[i+1]
                i += 2

    def save(self, save_name: str = "weights"):
        weights = []
        for layer in self.layers:
            for neuron in layer.neurons:
                weights.append([i.data for i in neuron.w])
                weights.append(neuron.b.data)
        file_name = f'{save_name}.pkl'
        with open(file_name, 'wb') as f:
            pickle.dump(weights, f)



if __name__ == "__main__":

    import numpy as np
    import matplotlib.pyplot as plt
    import pandas as pd
    from sklearn.metrics import r2_score
    df_x = pd.read_csv("../../test/data/x.csv", sep=";")
    xs = df_x.to_numpy()
    print(xs.shape)
    df_y = pd.read_csv("../../test/data/x.csv/y.csv", sep=";")
    y = df_y.to_numpy().squeeze()
    print(y.shape)
    model = MLP(6, [16, 1], bias=True) # 1-layer neural network
    print(model)
    print("number of parameters", len(model.parameters()))

    # loop
    for k in range(30):
        for i in range(len(xs)):
            output = model(xs[i])
            target = y[i]
            loss = ((output - target) ** 2)
            loss.backward()
            for p in model.parameters():
                p.data += -0.001 * p.grad 
            model.zero_grad()
        if k%5 == 0:
            print(k, loss)
    
    print('\nTEST')
    output = model(xs[0])
    target = y[0]
    print(f'output: {output.data}', f'target: {target}')

    # r2 score
    output = [model(x).data for x in xs]
    r2 = r2_score(y, output)
    print(f'r2 score: {r2}')

